// Copyright 2018 Canonical Ltd.
// Licensed under the AGPLv3, see LICENCE file for details.

package apiserver

import (
	"time"

	"github.com/juju/errors"

	"github.com/juju/juju/core/leadership"
	"github.com/juju/juju/core/lease"
	"github.com/juju/juju/state"
)

// leadershipChecker implements leadership.Checker by wrapping a lease.Checker.
type leadershipChecker struct {
	checker lease.Checker
}

// LeadershipCheck is part of the leadership.Checker interface.
func (m leadershipChecker) LeadershipCheck(applicationName, unitName string) leadership.Token {
	token := m.checker.Token(applicationName, unitName)
	return leadershipToken{
		applicationName: applicationName,
		unitName:        unitName,
		token:           token,
	}
}

// leadershipToken implements leadership.Token by wrapping a lease.Token.
type leadershipToken struct {
	applicationName string
	unitName        string
	token           lease.Token
}

// Check is part of the leadership.Token interface.
func (t leadershipToken) Check(attempt int, out interface{}) error {
	err := t.token.Check(attempt, out)
	if errors.Cause(err) == lease.ErrNotHeld {
		return errors.Errorf("%q is not leader of %q", t.unitName, t.applicationName)
	}
	return errors.Trace(err)
}

// leadershipClaimer implements leadership.Claimer by wrapping a lease.Claimer.
type leadershipClaimer struct {
	claimer lease.Claimer
}

// ClaimLeadership is part of the leadership.Claimer interface.
func (m leadershipClaimer) ClaimLeadership(applicationName, unitName string, duration time.Duration) error {
	err := m.claimer.Claim(applicationName, unitName, duration)
	if errors.Cause(err) == lease.ErrClaimDenied {
		return leadership.ErrClaimDenied
	}
	return errors.Trace(err)
}

// BlockUntilLeadershipReleased is part of the leadership.Claimer interface.
func (m leadershipClaimer) BlockUntilLeadershipReleased(applicationName string, cancel <-chan struct{}) error {
	err := m.claimer.WaitUntilExpired(applicationName, cancel)
	if errors.Cause(err) == lease.ErrWaitCancelled {
		return leadership.ErrBlockCancelled
	}
	return errors.Trace(err)
}

// leadershipPinner implements leadership.Pinner by wrapping a lease.Pinner.
type leadershipPinner struct {
	pinner lease.Pinner
}

// PinLeadership (leadership.Pinner) pins the lease
// for the input application and entity.
func (m leadershipPinner) PinLeadership(applicationName string, entity string) error {
	return errors.Trace(m.pinner.Pin(applicationName, entity))
}

// UnpinLeadership (leadership.Pinner) unpins the lease
// for the input application and entity.
func (m leadershipPinner) UnpinLeadership(applicationName string, entity string) error {
	return errors.Trace(m.pinner.Unpin(applicationName, entity))
}

// PinnedLeadership (leadership.Pinner) returns applications for which
// leadership is pinned, along with the entities requiring the
// pinned behaviour.
func (m leadershipPinner) PinnedLeadership() map[string][]string {
	return m.pinner.Pinned()
}

// leadershipReader implements leadership.Reader by wrapping a lease.Reader.
type leadershipReader struct {
	reader lease.Reader
}

// Leaders (leadership.Reader) returns all application leaders in the
// current model.
func (r leadershipReader) Leaders() (map[string]string, error) {
	return r.reader.Leases(), nil
}

// legacyLeadershipReader implements leadership.Reader by wrapping state.
type legacyLeadershipReader struct {
	st *state.State
}

// Leaders (leadership.Reader) returns all application leaders in the
// current model according to state.
func (r legacyLeadershipReader) Leaders() (map[string]string, error) {
	l, err := r.st.ApplicationLeaders()
	return l, errors.Trace(err)
}
